#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>

#include <rlcsa/misc/utils.h>

#include "pdltree.h"


int
main(int argc, char** argv)
{
  std::cout << "Building PDL tree and sets" << std::endl;
  if(argc < 5)
  {
    std::cout << "Usage: build_pdltree mode base_name block_size storing_factor" << std::endl;
    std::cout << "  -c  PDL-count (storing_factor is not used)" << std::endl;
    std::cout << "  -f  PDL-fast (block_size and storing_factor are not used)" << std::endl;
    std::cout << "  -m  PDL-topk with merging (storing_factor is used)" << std::endl;
    std::cout << "  -r  PDL-RP (default)" << std::endl;
    std::cout << "  -s  PDL-set" << std::endl;
    std::cout << "  -t  PDL-topk (storing_factor is not used)" << std::endl;
    return 1;
  }

  PDLTree::mode_type mode = PDLTree::mode_rp;
  if(argv[1][0] == '-')
  {
    switch(argv[1][1])
    {
      case 'c':
        std::cout << "Mode: count" << std::endl;
        mode = PDLTree::mode_count; break;
      case 'f':
        std::cout << "Mode: fast" << std::endl;
        mode = PDLTree::mode_fc; break;
      case 'm':
        std::cout << "Mode: merge" << std::endl;
        mode = PDLTree::mode_merge; break;
      case 'r':
        std::cout << "Mode: rp" << std::endl;
        mode = PDLTree::mode_rp; break;
      case 's':
        std::cout << "Mode: set" << std::endl;
        mode = PDLTree::mode_set; break;
      case 't':
        std::cout << "Mode: topk" << std::endl;
        mode = PDLTree::mode_topk; break;
      default:
        std::cerr << "build_pdltree: Invalid mode: " << argv[1] << "!" << std::endl;
        break;
    }
  }

  std::string base_name = argv[2];
  std::cout << "Base name: " << base_name << std::endl;
  usint block_size = atoi(argv[3]);
  std::cout << "Block size: " << block_size << std::endl;
  usint storing_factor = atoi(argv[4]);
  std::cout << "Storing factor: " << storing_factor << std::endl;
  std::cout << std::endl;

  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { return 2; }
  rlcsa.printInfo();
  rlcsa.reportSize(true);
  usint size = rlcsa.getSize();

  double start = CSA::readTimer();
  PDLTree pdltree(rlcsa, block_size, storing_factor, mode, true);
  if(!(pdltree.isOk())) { return 3; }
  double total = CSA::readTimer() - start;

  std::cout << "Writing tree..." << std::endl;
  std::string pdltree_name = base_name + pdltree.getExtension();
  std::ofstream pdltree_file(pdltree_name.c_str(), std::ios_base::binary);
  if(!pdltree_file)
  {
    std::cerr << "build_pdltree: Cannot open output file " << pdltree_name << "!" << std::endl;
    return 4;
  }
  pdltree.writeTo(pdltree_file);

  usint freq_bytes = 0;
  if(pdltree.usesFrequencies())
  {
    std::cout << "Writing frequencies..." << std::endl;
    CSA::DeltaMultiArray* freqs = pdltree.getFrequencies(true);
    if(freqs != 0)
    {
      freq_bytes = freqs->reportSize();
      freqs->writeTo(pdltree_file);
      delete freqs; freqs = 0;
    }
  }

  usint count_bytes = 0;
  if(pdltree.usesCounts())
  {
    std::cout << "Writing document counts..." << std::endl;
    CSA::Array* counts = pdltree.getCounts();
    if(counts != 0)
    {
      count_bytes = counts->reportSize();
      counts->writeTo(pdltree_file);
      delete counts; counts = 0;
    }
  }
  pdltree_file.close();

  if(pdltree.usesSets())
  {
    std::cout << "Writing sets..." << std::endl;
    std::string set_name = base_name + PDLTree::SETS_EXTENSION;
    std::ofstream set_file(set_name.c_str(), std::ios_base::binary);
    if(!set_file)
    {
      std::cerr << "build_pdltree: Cannot open output file " << set_name << "!" << std::endl;
      return 6;
    }
    pdltree.writeSets(set_file);
    set_file.close();
    std::cout << std::endl;
  }
  std::cout << std::endl;

  double megabytes = size / (double)CSA::MEGABYTE;
  double pdl_size = pdltree.reportSize() / (double)CSA::MEGABYTE;
  double freq_size = freq_bytes / (double)CSA::MEGABYTE;
  double count_size = count_bytes / (double)CSA::MEGABYTE;
  double total_size = pdl_size + freq_size + count_size;
  std::cout << "Processed " << megabytes << " megabytes in " << total << " seconds (" << (megabytes / total) << " MB/s)." << std::endl;
  std::cout << "Memory: " << CSA::memoryUsage() << " kB" << std::endl;
  std::cout << "PDLTree: " << pdl_size << " MB (" << (CHAR_BIT * pdl_size / megabytes) << " bpc)" << std::endl;
  if(freq_bytes > 0)
  {
    std::cout << "Freqs: " << freq_size << " MB (" << (CHAR_BIT * freq_size / megabytes) << " bpc)" << std::endl;
  }
  if(count_bytes > 0)
  {
    std::cout << "Counts: " << count_size << " MB (" << (CHAR_BIT * count_size / megabytes) << " bpc)" << std::endl;
  }
  std::cout << "Total: " << total_size << " MB (" << (CHAR_BIT * total_size / megabytes) << " bpc)" << std::endl;
  std::cout << std::endl;

  return 0;
}
