#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <vector>

#include "pdlrp.h"
#include "pdlcount.h"
#include "sadacount.h"
#include "utils.h"

#include <rlcsa/misc/utils.h>


void printUsage();


int
main(int argc, char** argv)
{
  std::cout << "Document counting test" << std::endl;
  if(argc < 3)
  {
    printUsage();
    return 1;
  }

  // Parameters
  char* base_arg = 0;
  char* pattern_arg = 0;
  bool brute = false, brute_d = false, pdlcount = false, pdlfast = false, pdlrp = false, sada = false, selected = false;
  for(int arg = 1; arg < argc; arg++)
  {
    if(argv[arg][0] == '-')
    {
      switch(argv[arg][1])
      {
        case 'b':
          brute = true; selected = true; break;
        case 'B':
          brute_d = true; selected = true; break;
        case 'c':
          pdlcount = true; selected = true; break;
        case 'f':
          pdlfast = true; selected = true; break;
        case 'r':
          pdlrp = true; selected = true; break;
        case 's':
          sada = true; selected = true; break;
        default:
          std::cerr << "Invalid parameter: " << argv[arg] << std::endl;
          break;
      }
    }
    else if(base_arg == 0)    { base_arg = argv[arg]; }
    else if(pattern_arg == 0) { pattern_arg = argv[arg]; }
  }
  if(pattern_arg == 0)
  {
    printUsage();
    return 1;
  }
  if(!selected) { brute = true; }

  // Output
  std::cout << "Algorithms:";
  if(brute) { std::cout << " brute-l"; }
  if(brute_d) { std::cout << " brute-d"; }
  if(pdlcount) { std::cout << " pdlcount"; }
  if(pdlfast) { std::cout << " pdlfast"; }
  if(pdlrp) { std::cout << " pdlrp"; }
  if(sada) { std::cout << "  sada"; }
  std::cout << std::endl;
  std::string base_name = base_arg;
  std::cout << "Base name: " << base_name << std::endl;
  std::string patterns_name = pattern_arg;
  std::cout << "Patterns: " << patterns_name << std::endl;
  std::cout << std::endl;

  // Index
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { return 2; }
  rlcsa.printInfo();
  rlcsa.reportSize(true);

  // Patterns
  std::ifstream pattern_file(patterns_name.c_str(), std::ios_base::binary);
  if(!pattern_file)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 3;
  }
  std::vector<std::string> rows;
  CSA::readRows(pattern_file, rows, true);
  pattern_file.close();

  // find()
  std::vector<pair_type> ranges; ranges.reserve(rows.size());
  usint total_size = 0, total_occ = 0;
  double start = CSA::readTimer();
  for(usint i = 0; i < rows.size(); i++)
  {
    pair_type result = rlcsa.count(rows[i]);
    total_size += rows[i].length(); total_occ += CSA::length(result);
    ranges.push_back(result);
  }
  double seconds = CSA::readTimer() - start;
  double megabytes = total_size / (double)CSA::MEGABYTE;
  std::cout << "Patterns:     " << rows.size() << " (" << megabytes << " MB)" << std::endl;
  std::cout << "Occurrences:  " << total_occ << std::endl;
  std::cout << "Find:         " << seconds << " seconds (" << (megabytes / seconds) << " MB/s, "
                                << (rows.size() / seconds) << " patterns/s)" << std::endl;
  std::cout << std::endl;

  // Brute force
  if(brute)
  {
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++) { docc += bruteForceDocCount(rlcsa, ranges[i]); }
    seconds = CSA::readTimer() - start;
    printResults("Brute-L", seconds, docc, rows.size());
    printSize("Brute-L", rlcsa.getSize(), rlcsa.reportSize(), 0);
    std::cout << std::endl;
  }

  // Brute force with DA
  if(brute_d)
  {
    CSA::ReadBuffer* docarray = readDocarray(rlcsa, base_name);
    if(docarray == 0)
    {
      std::cerr << "count_docs: Cannot initialize DoclistSada!" << std::endl;
      return 4;
    }
    double start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++) { docc += bruteForceDocCount(*docarray, ranges[i]); }
    double seconds = CSA::readTimer() - start;
    printResults("Brute-D", seconds, docc, ranges.size());
    printSize("Brute-D", rlcsa.getSize(), rlcsa.reportSize(), docarray->reportSize());
    std::cout << std::endl;
    delete docarray; docarray = 0;
  }

  if(pdlcount)
  {
    PDLCount pdlcount(rlcsa, base_name, false);
    if(!(pdlcount.isOk()))
    {
      std::cerr << "count_docs: Cannot initialize PDL-count!" << std::endl;
      return 5;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++) { docc += pdlcount.count(ranges[i]); }
    seconds = CSA::readTimer() - start;
    printResults("PDL-count", seconds, docc, rows.size());
    printSize("PDL-count", rlcsa.getSize(), rlcsa.reportSize(), pdlcount.reportSize());
    std::cout << std::endl;
  }

  if(pdlfast)
  {
    PDLCount pdlcount(rlcsa, base_name, true);
    if(!(pdlcount.isOk()))
    {
      std::cerr << "count_docs: Cannot initialize PDL-fast!" << std::endl;
      return 6;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++) { docc += pdlcount.count(ranges[i]); }
    seconds = CSA::readTimer() - start;
    printResults("PDL-fast", seconds, docc, rows.size());
    printSize("PDL-fast", rlcsa.getSize(), rlcsa.reportSize(), pdlcount.reportSize());
    std::cout << std::endl;
  }

  if(pdlrp)
  {
    PDLRP rp(rlcsa, base_name, false);
    if(!(rp.isOk()))
    {
      std::cerr << "count_docs: Cannot initialize PDL-RP!" << std::endl;
      return 7;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++) { docc += rp.count(ranges[i]); }
    seconds = CSA::readTimer() - start;
    printResults("PDL-RP", seconds, docc, rows.size());
    printSize("PDL-RP", rlcsa.getSize(), rlcsa.reportSize(), rp.reportSize());
    std::cout << std::endl;
  }

  if(sada)
  {
    SadaCount sadacount(rlcsa, base_name);
    if(!(sadacount.isOk()))
    {
      std::cerr << "count_docs: Cannot initialize SadaCount!" << std::endl;
      return 8;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++) { docc += sadacount.count(ranges[i]); }
    seconds = CSA::readTimer() - start;
    printResults("Sada", seconds, docc, rows.size());
    printSize("Sada", rlcsa.getSize(), rlcsa.reportSize(), sadacount.reportSize());
    std::cout << std::endl;
  }

  return 0;
}

void
printUsage()
{
  std::cout << "Usage: count_docs [algorithms] base_name patterns" << std::endl;
  std::cout << "  -b  Brute force using locate() (default)" << std::endl;
  std::cout << "  -B  Brute force using D" << std::endl;
  std::cout << "  -c  Precomputed document counting" << std::endl;
  std::cout << "  -f  Fast precomputed document counting" << std::endl;
  std::cout << "  -r  Precomputed document listing using RePair" << std::endl;
  std::cout << "  -s  Sadakane's document counting using RLE" << std::endl;
  std::cout << std::endl;
}
