#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <vector>

#include "doclist.h"
#include "pdlrp.h"
#include "utils.h"

#include <rlcsa/misc/utils.h>
#include <rlcsa/docarray.h>


void printUsage();
void writeRanges(std::vector<pair_type>& ranges, const std::string& file_name);
void listDocuments(const Doclist& doclist, const RLCSA& rlcsa, const std::vector<pair_type>& ranges, const std::string& algorithm);

int
main(int argc, char** argv)
{
  std::cout << "Document listing test" << std::endl;
  if(argc < 3)
  {
    printUsage();
    return 1;
  }

  // Parameters
  char* base_arg = 0;
  char* pattern_arg = 0;
  bool brute = false, brute_d = false, ilcp = false, ilcp_muthu = false, pdl = false, pdlrp = false, pdlset = false, sada = false, sada_muthu = false, selected = false;
  for(int arg = 1; arg < argc; arg++)
  {
    if(argv[arg][0] == '-')
    {
      switch(argv[arg][1])
      {
        case 'b':
          brute = true; selected = true; break;
        case 'B':
          brute_d = true; selected = true; break;
        case 'i':
          ilcp = true; selected = true; break;
        case 'I':
          ilcp_muthu = true; selected = true; break;
        case 'p':
          pdl = true; selected = true; break;
        case 'r':
          pdlrp = true; selected = true; break;
        case 'R':
          pdlset = true; selected = true; break;
        case 's':
          sada = true; selected = true; break;
        case 'S':
          sada_muthu = true; selected = true; break;
        default:
          std::cerr << "Invalid parameter: " << argv[arg] << std::endl;
          break;
      }
    }
    else if(base_arg == 0)    { base_arg = argv[arg]; }
    else if(pattern_arg == 0) { pattern_arg = argv[arg]; }
  }
  if(pattern_arg == 0)
  {
    printUsage();
    return 1;
  }
  if(!selected) { brute = true; }

  // Output
  std::cout << "Algorithms:";
  if(brute) { std::cout << " brute-l"; }
  if(brute_d) { std::cout << " brute-d"; }
  if(ilcp)  { std::cout << " ilcp-l"; }
  if(ilcp_muthu)  { std::cout << " ilcp-d"; }
  if(pdl)  { std::cout << " pdl"; }
  if(pdlrp) { std::cout << " pdlrp"; }
  if(pdlset) { std::cout << " pdlset"; }
  if(sada)  { std::cout << " sada-l"; }
  if(sada_muthu)  { std::cout << " sada-d"; }
  std::cout << std::endl;
  std::string base_name = base_arg;
  std::cout << "Base name: " << base_name << std::endl;
  std::string patterns_name = pattern_arg;
  std::cout << "Patterns: " << patterns_name << std::endl;
  std::cout << std::endl;

  // Index
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { return 2; }
  rlcsa.printInfo();
  rlcsa.reportSize(true);

  // Patterns
  std::ifstream pattern_file(patterns_name.c_str(), std::ios_base::binary);
  if(!pattern_file)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 3;
  }
  std::vector<std::string> rows;
  CSA::readRows(pattern_file, rows, true);
  pattern_file.close();

  // Counting
  std::vector<pair_type> ranges; ranges.reserve(rows.size());
  usint total_size = 0, total_occ = 0;
  double start = CSA::readTimer();
  for(usint i = 0; i < rows.size(); i++)
  {
    pair_type result = rlcsa.count(rows[i]);
    total_size += rows[i].length(); total_occ += CSA::length(result);
    ranges.push_back(result);
  }
  double seconds = CSA::readTimer() - start;
  double megabytes = total_size / (double)CSA::MEGABYTE;
  std::cout << "Patterns:     " << rows.size() << " (" << megabytes << " MB)" << std::endl;
  std::cout << "Occurrences:  " << total_occ << std::endl;
  std::cout << "Counting:     " << seconds << " seconds (" << (megabytes / seconds) << " MB/s, "
                                << (rows.size() / seconds) << " patterns/s)" << std::endl;
  std::cout << std::endl;
  writeRanges(ranges, patterns_name + ".ranges");

  // Brute force
  if(brute)
  {
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      usint* docs = rlcsa.locate(ranges[i]);
      if(docs != 0)
      {
        usint temp = CSA::length(ranges[i]);
        rlcsa.getSequenceForPosition(docs, temp);
        CSA::sequentialSort(docs, docs + temp);
        docc += std::unique(docs, docs + temp) - docs;
        delete docs; docs = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("Brute-L", seconds, docc, rows.size());
    printSize("Brute-L", rlcsa.getSize(), rlcsa.reportSize(), 0);
    std::cout << std::endl;
  }

  // Brute force with DA
  if(brute_d)
  {
    DoclistSada doclist(rlcsa, base_name, true);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistSada!" << std::endl;
      return 4;
    }
    double start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < ranges.size(); i++)
    {
      Doclist::result_type* res = doclist.listDocumentsBrute(ranges[i]);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    double seconds = CSA::readTimer() - start;
    printResults("Brute-D", seconds, docc, ranges.size());
    printSize("Brute-D", rlcsa.getSize(), rlcsa.reportSize(), doclist.reportSizeBrute());
    std::cout << std::endl;
  }

  // ILCP
  if(ilcp)
  {
    DoclistILCP doclist(rlcsa, base_name);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistILCP!" << std::endl;
      return 5;
    }
    listDocuments(doclist, rlcsa, ranges, "ILCP-L");
  }

  // ILCP
  if(ilcp_muthu)
  {
    DoclistILCP doclist(rlcsa, base_name, true);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistILCP!" << std::endl;
      return 6;
    }
    listDocuments(doclist, rlcsa, ranges, "ILCP-D");
  }

  if(pdl)
  {
    CSA::DocArray doclist(rlcsa, base_name);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DocArray!" << std::endl;
      return 7;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      std::vector<usint>* res = doclist.listDocuments(ranges[i]);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("PDL", seconds, docc, rows.size());
    printSize("PDL", rlcsa.getSize(), rlcsa.reportSize(), doclist.reportSize());
    std::cout << std::endl;
  }

  if(pdlrp)
  {
    PDLRP rp(rlcsa, base_name, false);
    if(!(rp.isOk()))
    {
      std::cerr << "Error: Cannot initialize PDLRP!" << std::endl;
      return 8;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      PDLRP::result_type* res = rp.query(ranges[i]);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("PDL-RP", seconds, docc, rows.size());
    printSize("PDL-RP", rlcsa.getSize(), rlcsa.reportSize(), rp.reportSize());
    std::cout << std::endl;
  }

  if(pdlset)
  {
    PDLRP rp(rlcsa, base_name, false, true);
    if(!(rp.isOk()))
    {
      std::cerr << "Error: Cannot initialize PDL-set!" << std::endl;
      return 9;
    }
    start = CSA::readTimer();
    usint docc = 0;
    for(usint i = 0; i < rows.size(); i++)
    {
      PDLRP::result_type* res = rp.query(ranges[i]);
      if(res != 0)
      {
        docc += res->size();
        delete res; res = 0;
      }
    }
    seconds = CSA::readTimer() - start;
    printResults("PDL-set", seconds, docc, rows.size());
    printSize("PDL-set", rlcsa.getSize(), rlcsa.reportSize(), rp.reportSize());
    std::cout << std::endl;
  }

  // Sada
  if(sada)
  {
    DoclistSada doclist(rlcsa, base_name);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistSada!" << std::endl;
      return 10;
    }
    listDocuments(doclist, rlcsa, ranges, "Sada-L");
  }

  // Sada-Muthu
  if(sada_muthu)
  {
    DoclistSada doclist(rlcsa, base_name, true);
    if(!(doclist.isOk()))
    {
      std::cerr << "Error: Cannot initialize DoclistSada!" << std::endl;
      return 11;
    }
    listDocuments(doclist, rlcsa, ranges, "Sada-D");
  }

  return 0;
}

void
printUsage()
{
  std::cout << "Usage: list_docs [algorithms] base_name patterns" << std::endl;
  std::cout << "  -b  Brute force using locate() (default)" << std::endl;
  std::cout << "  -B  Brute force using D" << std::endl;
  std::cout << "  -i  RMQ over ILCP using locate()" << std::endl;
  std::cout << "  -I  RMQ over ILCP using D" << std::endl;
  std::cout << "  -p  Precomputed document listing" << std::endl;
  std::cout << "  -r  Precomputed document listing using RePair" << std::endl;
  std::cout << "  -R  Precomputed document listing using RePair for sets" << std::endl;
  std::cout << "  -s  RMQ over C using locate()" << std::endl;
  std::cout << "  -S  RMQ over C using D" << std::endl;
  std::cout << std::endl;
}

void
writeRanges(std::vector<pair_type>& ranges, const std::string& file_name)
{
  if(ranges.empty()) { return; }

  std::ofstream output(file_name.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "Error: Cannot open file " << file_name << " for writing patterns!" << std::endl;
    return;
  }

  for(std::vector<pair_type>::iterator iter = ranges.begin(); iter != ranges.end(); ++iter)
  {
    uint temp = iter->first;
    output.write((char*)&temp, sizeof(temp));
    temp = iter->second;
    output.write((char*)&temp, sizeof(temp));
  }
  output.close();
}

void
listDocuments(const Doclist& doclist, const RLCSA& rlcsa, const std::vector<pair_type>& ranges, const std::string& algorithm)
{
  double start = CSA::readTimer();
  Doclist::found_type found(rlcsa.getNumberOfSequences(), 1);
  usint docc = 0;
  for(usint i = 0; i < ranges.size(); i++)
  {
    Doclist::result_type* res = doclist.listDocuments(ranges[i], &found);
    if(res != 0)
    {
      docc += res->size();
      delete res; res = 0;
    }
  }
  double seconds = CSA::readTimer() - start;
  printResults(algorithm, seconds, docc, ranges.size());
  printSize(algorithm, rlcsa.getSize(), rlcsa.reportSize(), doclist.reportSize());
  std::cout << std::endl;
}
