#include <algorithm>
#include <fstream>
#include <iostream>
#include <stack>

#include "pdltopk.h"

//--------------------------------------------------------------------------

const std::string PDLTopk::EXTENSION = ".topk";
const std::string PDLTopk::MERGE_EXTENSION = ".topk_merge";

PDLTopk::PDLTopk(const RLCSA& _rlcsa, const std::string& base_name, bool compress_sets, bool merge) :
  rlcsa(_rlcsa), status(st_error), supports_merge(merge),
  tree(0), freqs(0), grammar(0), blocks(0)
{
  if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
  {
    std::cerr << "PDLTopk::PDLTopk(): Invalid RLCSA!" << std::endl;
    return;
  }

  std::string input_name = base_name;
  if(compress_sets) { input_name += PDLTree::EXTENSION; }
  else if(this->supportsMerge()) { input_name += MERGE_EXTENSION; }
  else { input_name += EXTENSION; }
  std::ifstream input(input_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "PDLTopk::PDLTopk(): Cannot open input file " << input_name << "!" << std::endl;
    return;
  }
  this->tree = new PDLTree(this->rlcsa, input);
  if(!(this->tree->isOk())) { return; }
  this->freqs = CSA::MultiArray::readFrom(input);
  if(!(this->freqs->isOk())) { return; }
  this->status = st_unfinished;

  if(compress_sets)
  {
    uint terminals = 0;
    this->grammar = readGrammar(*(this->tree), base_name, terminals);
    this->blocks = readBlocks(*(this->tree), base_name, terminals);
  }
  else
  {
    pair_type temp;
    input.read((char*)&temp, sizeof(temp)); // Items, bits.
    this->grammar = new CSA::ReadBuffer(input, temp.first, temp.second);
    this->blocks = CSA::MultiArray::readFrom(input);
  }
  input.close();
  if(this->blocks == 0 || !(this->blocks->isOk())) { return; }

  this->status = st_ok;
}

PDLTopk::~PDLTopk()
{
  delete this->tree; this->tree = 0;
  delete this->freqs; this->freqs = 0;
  delete this->grammar; this->grammar = 0;
  delete this->blocks; this->blocks = 0;
}

void
PDLTopk::writeTo(const std::string& base_name) const
{
  if(!(this->isOk()))
  {
    std::cerr << "PDLTopk::writeTo(): Status is not ok!" << std::endl;
    return;
  }

  std::string filename = base_name + (this->supportsMerge() ? MERGE_EXTENSION : EXTENSION);
  std::ofstream output(filename.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "PDLTopk::writeTo(): Cannot open output file " << filename << "!" << std::endl;
    return;
  }

  this->tree->writeTo(output);
  this->freqs->writeTo(output);

  pair_type temp(this->grammar->getNumberOfItems(), this->grammar->getItemSize());
  output.write((char*)&temp, sizeof(temp));
  this->grammar->writeTo(output);
  this->blocks->writeTo(output);

  output.close();
}

usint
PDLTopk::reportSize(bool include_freqs) const
{
  usint bytes = sizeof(*this) + this->tree->reportSize() + this->grammar->reportSize() + this->blocks->reportSize();
  if(include_freqs) { bytes += this->freqs->reportSize(); }
  return bytes;
}

//--------------------------------------------------------------------------

PDLTopk::result_type*
PDLTopk::query(const std::string& pattern, usint k, bool get_freqs) const
{
  if(!(this->isOk())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return 0; }
  return this->queryUnsafe(sa_range, k, get_freqs);
}

PDLTopk::result_type*
PDLTopk::query(pair_type sa_range, usint k, bool get_freqs) const
{
  if(!(this->isOk()) || CSA::isEmpty(sa_range) || sa_range.second >= this->rlcsa.getSize()) { return 0; }
  return this->queryUnsafe(sa_range, k, get_freqs);
}

PDLTopk::result_type*
PDLTopk::queryUnsafe(pair_type sa_range, usint k, bool get_freqs) const
{
  usint block = this->tree->getBlock(sa_range);
  if(block >= this->tree->getNumberOfNodes())
  {
    return bruteForceTopk(this->rlcsa, sa_range, k);
  }

  result_type* results = new result_type;
  this->addBlock(results, block, k, get_freqs);
  return results;
}

void
PDLTopk::addBlock(result_type* results, usint block, usint k, bool get_freqs) const
{
  if(k == 0) { k = CSA::WORD_MAX; }

  CSA::MultiArray::Iterator* block_iter = this->blocks->getIterator();
  block_iter->goToItem(block, 0); block_iter->setEnd(block + 1, 0);

  CSA::MultiArray::Iterator* freq_iter = 0;
  if(get_freqs) { freq_iter = this->freqs->getIterator(); freq_iter->goToItem(block, 0); }

  usint freq = 0, run = 0;
  std::stack<usint> buffer;
  while(results->size() < k && !(block_iter->atEnd()))
  {
    buffer.push(block_iter->nextItem());
    while(results->size() < k && !(buffer.empty()))
    {
      usint value = buffer.top(); buffer.pop();
      if(this->tree->isTerminal(value))
      {
        if(get_freqs)
        {
          if(run == 0) { freq = freq_iter->nextItem(); run = freq_iter->nextItem(); }
          run--;
        }
        results->push_back(Document(value, freq));
      }
      else
      {
        value = this->tree->toRule(value);
        buffer.push(this->grammar->readItemConst(2 * value + 1));
        buffer.push(this->grammar->readItemConst(2 * value));
      }
    }
  }

  delete block_iter; block_iter = 0;
  delete freq_iter; freq_iter = 0;
}

//--------------------------------------------------------------------------

PDLTopk::result_type*
PDLTopk::query(const std::string& pattern, usint k, merge_type strategy) const
{
  if(!(this->isOk()) || !(this->supportsMerge())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return 0; }
  return this->queryUnsafe(sa_range, k, strategy);
}

PDLTopk::result_type*
PDLTopk::query(pair_type sa_range, usint k, merge_type strategy) const
{
  if(!(this->isOk()) || !(this->supportsMerge()) ||
     CSA::isEmpty(sa_range) || sa_range.second >= this->rlcsa.getSize())
  {
    return 0;
  }
  return this->queryUnsafe(sa_range, k, strategy);
}

struct IntermediateIdComparator
{
  bool operator() (const IntermediateDoc& a, const IntermediateDoc& b)
  {
    return (a.id < b.id);
  }
} ii_comparator;

struct IntermediateFreqComparator
{
  bool operator() (const IntermediateDoc& a, const IntermediateDoc& b)
  {
    // Descending by frequency, ascending by id.
    if(a.freq != b.freq) { return (a.freq > b.freq); }
    return (a.id < b.id);
  }
} if_comparator;

PDLTopk::result_type*
PDLTopk::queryUnsafe(pair_type sa_range, usint k, merge_type strategy) const
{
  std::vector<usint> blocks;
  this->tree->getBlocks(sa_range, blocks);
  if(blocks.empty())
  {
    return bruteForceTopk(this->rlcsa, sa_range, k);
  }

  result_type* results = new result_type;
  if(strategy == merge_brute)
  {
    for(usint i = 0; i < blocks.size(); i++) { this->addBlock(results, blocks[i], 0, true); }
    sortById(results, false);
    mergeDocumentRuns(results);
    if(results->size() > k) { results->resize(k); }
    return results;
  }

  intermediate_type doc_buffer;
  CSA::MultiArray::Iterator** block_iters = new CSA::MultiArray::Iterator*[blocks.size()];
  CSA::MultiArray::Iterator** freq_iters = new CSA::MultiArray::Iterator*[blocks.size()];
  std::vector<pair_type> next_freqs; next_freqs.reserve(blocks.size()); // (freq, run)
  std::stack<usint>* buffers = new std::stack<usint>[blocks.size()];
  std::vector<uint> cumulative_freqs(blocks.size() + 1, 0);
  for(usint i = 0; i < blocks.size(); i++)
  {
    block_iters[i] = this->blocks->getIterator();
    block_iters[i]->goToItem(blocks[i], 0); block_iters[i]->setEnd(blocks[i] + 1, 0);
    freq_iters[i] = this->freqs->getIterator(); freq_iters[i]->goToItem(blocks[i], 0);
    usint freq = freq_iters[i]->nextItem(); usint run = freq_iters[i]->nextItem();
    next_freqs.push_back(pair_type(freq, run));
  }

  usint amount = k;
  while(true)
  {
    bool finished =
      this->addFromBlocks(doc_buffer, blocks.size(), amount, block_iters, freq_iters, next_freqs, buffers);
    if(finished) { break; }

    // cumulative_freqs[i] is an upper bound for the total frequency for a document in the remaining
    // blocks, if it has been seen in i blocks.
    for(usint i = 0; i < blocks.size(); i++) { cumulative_freqs[i] = next_freqs[i].first; }
    CSA::sequentialSort(cumulative_freqs.begin(), cumulative_freqs.begin() + blocks.size());
    for(usint i = blocks.size(); i > 0; i--) { cumulative_freqs[i - 1] += cumulative_freqs[i]; }

    // Check whether one of the non-topk documents can still end up in top-k.
    uint max_estimate = cumulative_freqs[0];
    for(usint i = k; i < doc_buffer.size(); i++)
    {
      doc_buffer[i].updateEstimate(cumulative_freqs);
      max_estimate = std::max(max_estimate, doc_buffer[i].estimate);
    }
    if(doc_buffer.size() >= k && max_estimate <= doc_buffer[k - 1].freq) { break; }

    if(strategy == merge_exp) { amount *= 2; }
  }

  for(usint i = 0; i < blocks.size(); i++)
  {
    delete block_iters[i]; block_iters[i] = 0;
    delete freq_iters[i]; freq_iters[i] = 0;
  }
  delete[] block_iters; block_iters = 0;
  delete[] freq_iters; freq_iters = 0;
  delete[] buffers; buffers = 0;

  for(usint i = 0; i < k && i < doc_buffer.size(); i++) { results->push_back(doc_buffer[i].getDoc()); }
  return results;
}

bool
PDLTopk::addFromBlocks(intermediate_type& doc_buffer, usint blocks, usint amount,
  CSA::MultiArray::Iterator** block_iters, CSA::MultiArray::Iterator** freq_iters,
  std::vector<pair_type>& next_freqs, std::stack<usint>* buffers) const
{
  // Extract up to 'amount' documents from each of the blocks.
  usint total_added = 0;
  for(usint i = 0; i < blocks; i++)
  {
    usint added = 0;
    while(added < amount && !(block_iters[i]->atEnd() && buffers[i].empty()))
    {
      if(buffers[i].empty()) { buffers[i].push(block_iters[i]->nextItem()); }
      while(added < amount && !(buffers[i].empty()))
      {
        usint value = buffers[i].top(); buffers[i].pop();
        if(this->tree->isTerminal(value))
        {
          doc_buffer.push_back(IntermediateDoc(value, next_freqs[i].first, 1));
          added++; next_freqs[i].second--;
          if(next_freqs[i].second == 0)
          {
            if(block_iters[i]->atEnd() && buffers[i].empty()) { next_freqs[i] = pair_type(0, 0); }
            else
            {
              next_freqs[i].first = freq_iters[i]->nextItem();
              next_freqs[i].second = freq_iters[i]->nextItem();
            }
          }
        }
        else
        {
          value = this->tree->toRule(value);
          buffers[i].push(this->grammar->readItemConst(2 * value + 1));
          buffers[i].push(this->grammar->readItemConst(2 * value));
        }
      }
    }
    total_added += added;
  }

  // Do the actual merging.
  if(total_added > 0)
  {
    CSA::sequentialSort(doc_buffer.begin(), doc_buffer.end(), ii_comparator);
    intermediate_type temp_buffer;
    IntermediateDoc curr(0, 0, 0);
    for(intermediate_type::iterator iter = doc_buffer.begin(); iter != doc_buffer.end(); ++iter)
    {
      if(iter->id == curr.id) { curr.freq += iter->freq; curr.blocks += iter->blocks; }
      else
      {
        if(curr.freq > 0) { temp_buffer.push_back(curr); }
        curr = *iter;
      }
    }
    if(curr.freq > 0) { temp_buffer.push_back(curr); }
    doc_buffer.swap(temp_buffer);
    CSA::sequentialSort(doc_buffer.begin(), doc_buffer.end(), if_comparator);
  }

  return (total_added == 0);
}

//--------------------------------------------------------------------------
