#include <algorithm>
#include <fstream>
#include <iostream>
#include <stack>

#include <rlcsa/bits/vectors.h>

#include "sadacount.h"

//--------------------------------------------------------------------------

struct SadaTreeNode
{
  uint         string_depth;
  pair_type    range;
  usint        docc, redundant;

  SadaTreeNode* child;
  SadaTreeNode* sibling;

  std::vector<uint>* docs;

  SadaTreeNode(uint lcp, pair_type sa_range) :
    string_depth(lcp), range(sa_range),
    docc(0), redundant(0),
    child(0), sibling(0),
    docs(0)
  {
  }

  ~SadaTreeNode()
  {
    delete this->child; this->child = 0;
    while(this->sibling != 0)
    {
      SadaTreeNode* temp = this->sibling; this->sibling = this->sibling->sibling;
      temp->sibling = 0; delete temp;
    }
    delete this->docs; this->docs = 0;
  }

  void addChild(SadaTreeNode* node)
  {
    if(this->child == 0) { this->child = node; return; }
    this->child->addSibling(node);
  }

  void addSibling(SadaTreeNode* node)
  {
    if(this->sibling == 0) { this->sibling = node; return; }
    this->sibling->addSibling(node);
  }

  void addLeaves()
  {
    if(this->child == 0 && CSA::length(this->range) <= 1) { return; }

    usint expect = this->range.first;
    SadaTreeNode* prev = 0;
    for(SadaTreeNode* temp = this->child; temp != 0; temp = temp->sibling)
    {
      while(expect < temp->range.first) { prev = this->addLeaf(prev, temp, expect); expect++; }
      temp->addLeaves();
      expect = temp->range.second + 1;
      prev = temp;
    }

    while(expect < this->range.second + 1) { prev = this->addLeaf(prev, 0, expect); expect++; }
  }

  SadaTreeNode* addLeaf(SadaTreeNode* left, SadaTreeNode* right, usint pos)
  {
    SadaTreeNode* leaf = new SadaTreeNode(0, pair_type(pos, pos));
    if(left == 0) { this->child = leaf; }
    else          { left->sibling = leaf; }
    leaf->sibling = right;
    return leaf;
  }

  SadaTreeNode* parentOf(SadaTreeNode* left, SadaTreeNode* right)
  {
    SadaTreeNode* parent = new SadaTreeNode(this->string_depth, pair_type(left->range.first, right->range.second));
    parent->child = left; left->sibling = right; right->sibling = 0;
    return parent;
  }

  // Call addLeaves() first.
  void binarize()
  {
    usint children = 0;
    std::vector<SadaTreeNode*> array;
    for(SadaTreeNode* curr = this->child; curr != 0; curr = curr->sibling)
    {
      curr->binarize();
      array.push_back(curr); children++;
    }
    if(children <= 2) { return; }

    while(children > 2)
    {
      usint new_children = 0;
      for(usint i = 0; i < children / 2; i++)
      {
        array[new_children] = this->parentOf(array[2 * i], array[2 * i + 1]);
        new_children++;
      }
      if(2 * new_children < children) { array[new_children] = array[children - 1]; new_children++; }
      children = new_children;
    }
    this->child = array[0]; this->child->sibling = array[1];
  }

  void insert(SadaTreeNode* from)
  {
    this->docs->insert(this->docs->end(), from->docs->begin(), from->docs->end());
    delete from->docs; from->docs = 0;
  }

  const static usint PARALLEL_SORT_THRESHOLD = 4096;

  // Call binarize() first.
  void computeStoredDocuments(usint* documents)
  {
    this->docc = 0;
    delete this->docs; this->docs = 0;

    usint total_size = 0;
    if(this->child == 0)
    {
      total_size = CSA::length(this->range);
      this->docs = new std::vector<uint>; this->docs->reserve(total_size);
      for(usint i = this->range.first; i <= this->range.second; i++)
      {
        this->docs->push_back(documents[i]);
      }
    }
    else
    {
      total_size = this->child->docc + this->child->sibling->docc;
      this->docs = new std::vector<uint>; this->docs->reserve(total_size);
      this->insert(this->child); this->insert(this->child->sibling);
    }

    CSA::removeDuplicates(this->docs, (total_size >= PARALLEL_SORT_THRESHOLD));
    this->docc = this->docs->size(); this->redundant = total_size - this->docc;
  }

  usint determineSize()
  {
    usint size = 1;
    if(this->child != 0) { size += this->child->determineSize(); }
    if(this->sibling != 0) { size += this->sibling->determineSize(); }
    return size;
  }
};

usint
smallestEncoding(const CSA::ReadBuffer& raw_vec, usint block_size, usint n)
{
  std::vector<pair_type> sizes;

  CSA::RLEVector* rle_vec = CSA::encode<CSA::RLEVector>(raw_vec, block_size, n);
  sizes.push_back(pair_type(rle_vec->reportSize(), SadaCount::ENCODING_RLE));
  delete rle_vec; rle_vec = 0;

  CSA::DeltaVector* delta_vec = CSA::encode<CSA::DeltaVector>(raw_vec, block_size, n);
  sizes.push_back(pair_type(delta_vec->reportSize(), SadaCount::ENCODING_DELTA));
  delete delta_vec; delta_vec = 0;

  CSA::SuccinctVector* succinct_vec = CSA::encode<CSA::SuccinctVector>(raw_vec, block_size, n);
  sizes.push_back(pair_type(succinct_vec->reportSize(), SadaCount::ENCODING_SUCCINCT));
  delete succinct_vec; succinct_vec = 0;

  CSA::sequentialSort(sizes.begin(), sizes.end());
  return sizes[0].second;
}

CSA::BitVector*
buildSadaCount(const RLCSA& rlcsa, usint& encoding, bool print)
{
  if(!(rlcsa.isOk()) || !(rlcsa.supportsLocate()))
  {
    std::cerr << "buildSadaCount(): Invalid RLCSA!" << std::endl;
    return 0;
  }
  switch(encoding)
  {
    case SadaCount::ENCODING_SMALLEST:
    case SadaCount::ENCODING_RLE:
    case SadaCount::ENCODING_DELTA:
    case SadaCount::ENCODING_SUCCINCT:
      break;
    default:
      std::cerr << "buildSadaCount(): Invalid encoding: " << encoding << "!" << std::endl;
      return 0;
  }

  // Build SA.
  if(print) { std::cout << "Building SA..." << std::endl; }
  usint size = rlcsa.getSize();
  usint* documents = new usint[size];
  #pragma omp parallel for schedule(static)
  for(usint i = 0; i < size; i += CSA::MEGABYTE)
  {
    rlcsa.locate(pair_type(i, std::min(i + CSA::MEGABYTE - 1, size - 1)), documents + i);
  }

  // Build LCP.
  if(print) { std::cout << "Building LCP..." << std::endl; }
  uint* lcp = new uint[size + 1]; lcp[size] = 0;
  CSA::PLCPVector* plcpvec = rlcsa.buildPLCP(16);
  CSA::PLCPVector::Iterator iter(*plcpvec);
  for(usint i = 0; i < size; i++)
  {
    lcp[i] = iter.select(documents[i]) - 2 * documents[i];
  }
  delete plcpvec; plcpvec = 0;

  // Convert SA to DA.
  if(print) { std::cout << "Building DA..." << std::endl; }
  #pragma omp parallel for schedule(static)
  for(usint i = 0; i < size; i += CSA::MEGABYTE)
  {
    rlcsa.getSequenceForPosition(documents + i, std::min(CSA::MEGABYTE, size - i));
  }

  // Build ST. Kind of.
  if(print) { std::cout << "Building ST..."; std::cout.flush(); }
  std::stack<SadaTreeNode*> nodestack; nodestack.push(new SadaTreeNode(0, pair_type(0, 0)));
  SadaTreeNode* prev = 0;
  for(usint i = 1; i <= size; i++)
  {
    usint left = i - 1;
    while(lcp[i] < nodestack.top()->string_depth)
    {
      nodestack.top()->range.second = i - 1;
      prev = nodestack.top(); nodestack.pop();
      left = prev->range.first;
      if(lcp[i] <= nodestack.top()->string_depth) { nodestack.top()->addChild(prev); prev = 0; }
    }
    if(lcp[i] > nodestack.top()->string_depth)
    {
      SadaTreeNode* curr = new SadaTreeNode(lcp[i], pair_type(left, left));
      if(prev != 0) { curr->addChild(prev); prev = 0; }
      nodestack.push(curr);
    }
  }
  while(!(nodestack.empty()))
  {
    prev = nodestack.top(); nodestack.pop();
    prev->range.second = size - 1;
  }
  SadaTreeNode* root = prev;
  delete[] lcp; lcp = 0;
  root->addLeaves();
  root->binarize();
  if(print) { std::cout << " Size: " << root->determineSize() << " / " << size << std::endl; }

  // Build the sets of document ids.
  if(print) { std::cout << "Building sets..." << std::endl; }
  for(prev = root; prev != 0; prev = prev->child) { nodestack.push(prev); }
  while(!(nodestack.empty()))
  {
    SadaTreeNode* curr = nodestack.top(); nodestack.pop();
    for(SadaTreeNode* temp = curr->sibling; temp != 0; temp = temp->child) { nodestack.push(temp); }
    curr->computeStoredDocuments(documents);
  }
  delete[] documents; documents = 0;

  // Building the counting structure.
  if(print) { std::cout << "Building bitvector..." << std::endl; }
  CSA::WriteBuffer buffer(2 * (size / CSA::WORD_BITS + 1));
  usint bits = 0, ones = 0;
  while(!(nodestack.empty())) { prev = nodestack.top(); nodestack.pop(); }
  for(prev = root; prev != 0; prev = prev->child) { nodestack.push(prev); }
  while(!(nodestack.empty()))
  {
    SadaTreeNode* curr = nodestack.top(); nodestack.pop();
    if(curr->child != 0)
    {
      buffer.setBit(bits); bits += 1 + curr->redundant; ones++;
      for(prev = curr->child->sibling; prev != 0; prev = prev->child) { nodestack.push(prev); }
    }
  }
  delete root; root = 0;
  if(print)
  {
    std::cout << "  Vector: " << ones << " / " << bits << " bits" << std::endl;
  }
  CSA::ReadBuffer* raw_vec = buffer.getReadBuffer();

  // Choosing the encoding.
  CSA::BitVector* vec = 0;
  if(encoding == SadaCount::ENCODING_SMALLEST)
  {
    encoding = smallestEncoding(*raw_vec, SadaCount::BLOCK_SIZE, bits);
  }
  if(print) { std::cout << "  Encoding: "; }
  switch(encoding)
  {
    case SadaCount::ENCODING_RLE:
      if(print) { std::cout << "RLE"; }
      vec = CSA::encode<CSA::RLEVector>(*raw_vec, SadaCount::BLOCK_SIZE, bits);
      break;
    case SadaCount::ENCODING_DELTA:
      if(print) { std::cout << "Delta"; }
      vec = CSA::encode<CSA::DeltaVector>(*raw_vec, SadaCount::BLOCK_SIZE, bits);
      break;
    case SadaCount::ENCODING_SUCCINCT:
      if(print) { std::cout << "Succinct"; }
      vec = CSA::encode<CSA::SuccinctVector>(*raw_vec, SadaCount::BLOCK_SIZE, bits);
      break;
    default:
      if(print) { std::cout << "None"; }
      break;
  }
  if(print) { std::cout << std::endl << std::endl; }

  delete raw_vec; raw_vec = 0;
  return vec;
}

//--------------------------------------------------------------------------

const std::string SadaCount::EXTENSION = ".sadacount";

SadaCount::SadaCount(const RLCSA& _rlcsa, usint _encoding, bool print) :
  rlcsa(_rlcsa), status(st_error), encoding(_encoding),
  data_rle(0), data_delta(0), data_succinct(0)
{
  if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
  {
    std::cerr << "SadaCount::SadaCount(): Invalid RLCSA!" << std::endl;
    return;
  }

  CSA::BitVector* temp = buildSadaCount(this->rlcsa, this->encoding, print);
  switch(this->encoding)
  {
    case ENCODING_RLE:
      this->data_rle = (CSA::RLEVector*)temp;
      if(this->data_rle != 0) { this->status = st_ok; }
      break;
    case ENCODING_DELTA:
      this->data_delta = (CSA::DeltaVector*)temp;
      if(this->data_delta != 0) { this->status = st_ok; }
      break;
    case ENCODING_SUCCINCT:
      this->data_succinct = (CSA::SuccinctVector*)temp;
      if(this->data_succinct != 0) { this->status = st_ok; }
      break;
    default:
      delete temp; temp = 0; break;
  }
}

SadaCount::SadaCount(const RLCSA& _rlcsa, const std::string& base_name) :
  rlcsa(_rlcsa), status(st_error), encoding(ENCODING_SMALLEST),
  data_rle(0), data_delta(0), data_succinct(0)
{
  if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
  {
    std::cerr << "SadaCount::SadaCount(): Invalid RLCSA!" << std::endl;
    return;
  }

  std::string input_name = base_name + EXTENSION;
  std::ifstream input(input_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "SadaCount::SadaCount(): Cannot open input file " << input_name << "!" << std::endl;
    return;
  }

  usint flags = 0;
  input.read((char*)&flags, sizeof(flags));
  this->encoding = flags & ENCODING_MASK;
  switch(this->encoding)
  {
    case ENCODING_RLE:
      this->data_rle = new CSA::RLEVector(input); break;
    case ENCODING_DELTA:
      this->data_delta = new CSA::DeltaVector(input); break;
    case ENCODING_SUCCINCT:
      this->data_succinct = new CSA::SuccinctVector(input); break;
    default:
      std::cerr << "SadaCount::SadaCount(): Invalid encoding: " << this->encoding << "!" << std::endl;
      return;
  }
  input.close();

  this->status = st_ok;
}

SadaCount::~SadaCount()
{
  delete this->data_rle; this->data_rle = 0;
  delete this->data_delta; this->data_delta = 0;
  delete this->data_succinct; this->data_succinct = 0;
}

void
SadaCount::writeTo(const std::string& base_name) const
{
  if(!(this->isOk()))
  {
    std::cerr << "SadaCount::writeTo(): Status is not ok!" << std::endl;
    return;
  }

  std::string filename = base_name + EXTENSION;
  std::ofstream output(filename.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "SadaCount::writeTo(): Cannot open output file " << filename << "!" << std::endl;
    return;
  }

  usint flags = this->encoding;
  output.write((char*)&flags, sizeof(flags));
  switch(this->encoding)
  {
    case ENCODING_RLE:
      this->data_rle->writeTo(output); break;
    case ENCODING_DELTA:
      this->data_delta->writeTo(output); break;
    case ENCODING_SUCCINCT:
      this->data_succinct->writeTo(output); break;
  }
  output.close();
}

usint
SadaCount::reportSize() const
{
  usint bytes = sizeof(*this);
  if(this->data_rle != 0) { bytes += this->data_rle->reportSize(); }
  if(this->data_delta != 0) { bytes += this->data_delta->reportSize(); }
  if(this->data_succinct != 0) { bytes += this->data_succinct->reportSize(); }
  return bytes;
}

usint
SadaCount::countRuns() const
{
  if(this->data_rle != 0)
  {
    CSA::RLEVector::Iterator iter(*(this->data_rle));
    return iter.countRuns();
  }
  std::cerr << "SadaCount::countRuns(): Operation not supported with encoding " << this->encoding << "!" << std::endl;
  return 0;
}

//--------------------------------------------------------------------------

usint
SadaCount::count(const std::string& pattern) const
{
  if(!(this->isOk())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  return this->count(sa_range);
}

usint
SadaCount::count(pair_type sa_range) const
{
  if(!(this->isOk()) || sa_range.second >= this->rlcsa.getSize()) { return 0; }

  usint len = CSA::length(sa_range);
  if(len <= 1) { return len; }

  if(this->encoding == ENCODING_RLE)
  {
    CSA::RLEVector::Iterator iter(*(this->data_rle));
    return 2 * len - (iter.select(sa_range.second) - iter.select(sa_range.first)) - 1;
  }
  else if(this->encoding == ENCODING_DELTA)
  {
    CSA::DeltaVector::Iterator iter(*(this->data_delta));
    return 2 * len - (iter.select(sa_range.second) - iter.select(sa_range.first)) - 1;
  }
  else if(this->encoding == ENCODING_SUCCINCT)
  {
    CSA::SuccinctVector::Iterator iter(*(this->data_succinct));
    return 2 * len - (iter.select(sa_range.second) - iter.select(sa_range.first)) - 1;
  }
  return 0;
}

//--------------------------------------------------------------------------
